/* Crash Recovery 2006-03-07 - Copyleft © 2006  Simon Bünzli  <zeniko@gmail.com> */

/*const*/ var CrashRecoveryService = {
	mCID: Components.ID("{1280606b-2510-4fe0-97ef-9b5a22eafe6b}"),
	mContractID: "@zeniko/crashrecoveryservice;1",
	mClassName: "Crash Recovery Service",

	mObserverService: Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService),
	mPrefBranch: Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch2),
	mWindowMediator: Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator),
	mIOService: Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService),
	mUnicodeConverter: Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter),
	mProfileDirectory: null,

	mObserving: ["domwindowopened", "domwindowclosed", "quit-application-requested", "quit-application-granted", "quit-application", "browser:purge-session-history"],
	mPrefRoot: "extensions.crashrecovery.",

	mWinAttributes: ["width", "height", "screenX", "screenY", "sizemode"],
	mWinHidable: ["menubar", "toolbar", "locationbar", "personalbar", "statusbar", "scrollbars"],
	mCapabilities: ["Subframes", "Plugins", "Javascript", "MetaRedirects", "Images"],
	mXulAttributes: ["locked", "protected", "marked"],

	mLoadState: 0, // 1 = ready, -1 = shutting down
	mLastSaveTime: 0,
	mLastWindowClosed: null,
	mWindows: {},
	mDirty: {},

/* ........ Global Event Handlers .............. */

	init: function()
	{
		this.mObserving.forEach(function(aTopic) {
			CrashRecoveryService.mObserverService.addObserver(CrashRecoveryService, aTopic, false);
		});
		
		this.mPref_interval = this.getPref(this.mPrefRoot + "interval", 10000);
		this.mPref_postdata = this.getPref(this.mPrefRoot + "postdata", -1);
		this.mPref_privacy_level = this.getPref(this.mPrefRoot + "privacy_level", 1);
		this.mPref_resume_session = this.getPref(this.mPrefRoot + "resume_session", false);
		this.mPrefBranch.addObserver(this.mPrefRoot, this, false);
		
		this.mProfileDirectory = this._COM("@mozilla.org/file/directory_service;1", "nsIProperties").get("ProfD", Components.interfaces.nsILocalFile);
		this.mStringBundle = this.getStringBundle("chrome://crashrecovery/locale/crashrecovery.properties");
		this.mUnicodeConverter.charset = "UTF-8";
		this.initFileWriter();
		
		this.mInitialState = this.stateFromDisk();
		if (this.mInitialState)
		{
			this.stateToDisk(this.mInitialState, true, true);
			try
			{
				this.mInitialState = IniObjectSerializer.decode(this.mInitialState);
				this.mLastSessionCrashed = this.mInitialState.CrashRecovery && this.mInitialState.CrashRecovery.state && this.mInitialState.CrashRecovery.state == "running";
				
				if (this.mInitialState.Window && this.mInitialState.Window[0])
				{
					this.mWinAttributes.forEach(function(aAttr) {
						delete CrashRecoveryService.mInitialState.Window[0][aAttr];
					});
					delete this.mInitialState.Window[0].hidden;
				}
			}
			catch (ex) { debug(ex); }
		}
		if (this.mLastSessionCrashed)
		{
			try
			{
				this.restoreCache();
			}
			catch (ex) { debug(ex); }
		}
	},

	uninit: function()
	{
		if (this.doResumeSession())
		{
			this.saveState(true);
		}
		else
		{
			this.clearDisk();
		}
		
		this.mObserving.forEach(function(aTopic) {
			CrashRecoveryService.mObserverService.removeObserver(CrashRecoveryService, aTopic);
		});
		this.mPrefBranch.removeObserver(this.mPrefRoot, this);
	},

	observe: function(aSubject, aTopic, aData)
	{
		switch (aTopic)
		{
		case "app-startup":
			this.mObserverService.addObserver(this, "profile-after-change", false);
			break;
		case "profile-after-change":
			this.mObserverService.removeObserver(this, aTopic);
			this.init();
			break;
		case "domwindowopened":
			aSubject.addEventListener("load", this.onLoad_proxy, false);
			break;
		case "domwindowclosed":
			aSubject.removeEventListener("load", this.onLoad_proxy, false);
			this.onClose(aSubject, this.mQuitRequest && this.mQuitRequest > Date.now() - 3000);
			break;
		case "quit-application-requested":
			this.getBrowserWindows().forEach(function(aWindow) {
				CrashRecoveryService.saveWindowHistory(aWindow);
				CrashRecoveryService.updateTextAndFramesetData(aWindow);
				CrashRecoveryService.updateCookieHosts(aWindow);
				CrashRecoveryService.updateWindowFeatures(aWindow);
			});
			this.mDirty = {};
			this.mQuitRequest = Date.now();
			break;
		case "quit-application-granted":
		case "quit-application":
			if (this.mLoadState != -1)
			{
				this.mLoadState = -1;
				this.uninit();
			}
			break;
		case "browser:purge-session-history":
			this.getBrowserWindows().forEach(function(aWindow) {
				Array.forEach(aWindow.getBrowser().browsers, function(aBrowser) {
					delete aBrowser.__CR_data;
				});
			});
			this.clearDisk();
			this.saveState(true);
			break;
		case "nsPref:changed":
			var len = this.mPrefRoot.length;
			if (aData.substr(0, len) == this.mPrefRoot)
			{
				this["mPref_" + aData.substr(len)] = this.getPref(aData);
			}
			break;
		}
	},

/* ........ Window Event Handlers .............. */

	onLoad_proxy: function()
	{
		CrashRecoveryService.onLoad(this);
	},

	onLoad: function(aWindow)
	{
		if (aWindow.document.documentElement.getAttribute("windowtype") == "navigator:browser")
		{
			aWindow.__CRi = "window" + Date.now();
			this.mWindows[aWindow.__CRi] = { Tab: [], selected: 0 };
			
			if (this.mLoadState == 0)
			{
				this.mLoadState = 1;
				this.saveStateDelayed(aWindow, 3000);
				
				if (this.mInitialState && (this.mLastSessionCrashed)?this.doRecoverSession():this.doResumeSession())
				{
					this.mInitialState._firstTabs = true;
					this.restoreWindow(aWindow, this.mInitialState, this.isCmdLineEmpty(aWindow));
				}
				delete this.mInitialState;
				
				aWindow.setTimeout(this.retryDownloads, 0, aWindow);
			}
			
			var tabpanels = aWindow.getBrowser().mPanelContainer;
			for (var i = 0; i < tabpanels.childNodes.length; i++)
			{
				this.onTabAdd(aWindow, tabpanels.childNodes[i], true);
			}
			tabpanels.addEventListener("DOMNodeInserted", this.onTabAdd_proxy, false);
			tabpanels.addEventListener("DOMNodeRemoved", this.onTabRemove_proxy, false);
			tabpanels.addEventListener("select", this.onTabSelect_proxy, false);
		}
	},

	onClose: function(aWindow, aIgnoreClosure)
	{
		if (aWindow.document.documentElement.getAttribute("windowtype") == "navigator:browser")
		{
			var tabpanels = aWindow.getBrowser().mPanelContainer;
			for (var i = 0; i < tabpanels.childNodes.length; i++)
			{
				this.onTabRemove(aWindow, tabpanels.childNodes[i], true);
			}
			tabpanels.removeEventListener("DOMNodeInserted", this.onTabAdd_proxy, false);
			tabpanels.removeEventListener("DOMNodeRemoved", this.onTabRemove_proxy, false);
			tabpanels.removeEventListener("select", this.onTabSelect_proxy, false);
			
			if (this.mLoadState != -1 && !aIgnoreClosure)
			{
				this.saveWindowHistory(aWindow);
				this.updateTextAndFramesetData(aWindow);
				this.updateCookieHosts(aWindow);
				this.updateWindowFeatures(aWindow);
				
				this.mLastWindowClosed = this.mWindows[aWindow.__CRi];
				this.mLastWindowClosed.Cookies = this.getCookies(this.mLastWindowClosed._hosts);
				delete this.mWindows[aWindow.__CRi];
				delete aWindow.__CRi;
				
				var window = this.getBrowserWindows()[0] || null;
				if (window)
				{
					this.saveStateDelayed(window);
				}
				
				this.mObserverService.notifyObservers(aWindow, "crashrecovery:windowclosed", IniObjectSerializer.encode({ Window: [this.mLastWindowClosed] }));
			}
		}
	},

	onTabAdd_proxy: function(aEvent)
	{
		CrashRecoveryService.onTabAdd(this.ownerDocument.defaultView, aEvent.target);
	},

	onTabAdd: function(aWindow, aBrowser, aNoNotification)
	{
		aBrowser.addEventListener("load", this.onTabLoad_proxy, true);
		aBrowser.addEventListener("pageshow", this.onTabLoad_proxy, true);
		aBrowser.addEventListener("input", this.onTabInput_proxy, true);
		aBrowser.addEventListener("DOMAutoComplete", this.onTabInput_proxy, true);
		
		if (!aNoNotification)
		{
			this.saveStateDelayed(aWindow);
		}
	},

	onTabRemove_proxy: function(aEvent)
	{
		CrashRecoveryService.onTabRemove(this.ownerDocument.defaultView, aEvent.target);
	},

	onTabRemove: function(aWindow, aBrowser, aNoNotification)
	{
		aBrowser.removeEventListener("load", this.onTabLoad_proxy, true);
		aBrowser.removeEventListener("pageshow", this.onTabLoad_proxy, true);
		aBrowser.removeEventListener("input", this.onTabInput_proxy, true);
		aBrowser.removeEventListener("DOMAutoComplete", this.onTabInput_proxy, true);
		
		aBrowser = this.ensureBrowser(aBrowser);
		delete aBrowser.__CR_data;
		delete aBrowser.__CR_text;
		
		if (!aNoNotification)
		{
			this.saveStateDelayed(aWindow);
		}
	},

	onTabLoad_proxy: function(aEvent)
	{
		CrashRecoveryService.onTabLoad(this.ownerDocument.defaultView, this, aEvent);
	},

	onTabLoad: function(aWindow, aBrowser, aEvent)
	{
		if (aEvent.persisted == undefined || aEvent.persisted)
		{
			aBrowser = this.ensureBrowser(aBrowser);
			delete aBrowser.__CR_data;
			delete aBrowser.__CR_text;
			this.saveStateDelayed(aWindow);
		}
	},

	onTabInput_proxy: function(aEvent)
	{
		CrashRecoveryService.onTabInput(this.ownerDocument.defaultView, this, aEvent);
	},

	onTabInput: function(aWindow, aBrowser, aEvent)
	{
		this.saveTextData(this.ensureBrowser(aBrowser), XPCNativeWrapper(aEvent.originalTarget));
		this.saveStateDelayed(aWindow, 3000);
	},

	onTabSelect_proxy: function(aEvent)
	{
		CrashRecoveryService.onTabSelect(this.ownerDocument.defaultView, this);
	},

	onTabSelect: function(aWindow, aPanels)
	{
		if (this.mLoadState == 1)
		{
			this.mWindows[aWindow.__CRi].selected = aPanels.selectedIndex + 1;
			this.saveStateDelayed(aWindow);
		}
	},

/* ........ Saving Functionality .............. */

	saveWindowHistory: function(aWindow)
	{
		var tabbrowser = aWindow.getBrowser();
		var browsers = tabbrowser.browsers;
		if (!tabbrowser.mTabs) // compatibility fix for MultiZilla
		{
			tabbrowser.mTabs = tabbrowser.mTabContainer.childNodes;
		}
		var tabs = this.mWindows[aWindow.__CRi].Tab = [];
		this.mWindows[aWindow.__CRi].selected = 0;
		
		for (var i = 0; i < browsers.length; i++)
		{
			var browser = browsers[i];
			var history = null;
			var tabData = { Entry: [], index: 0 };
			
			try
			{
				history = browser.sessionHistory;
			}
			catch (ex) { }
			
			if (history && browser.__CR_data && browser.__CR_data.Entry[history.index])
			{
				tabData = browser.__CR_data;
				
				var content = XPCNativeWrapper(browser.contentWindow);
				tabData.Entry[history.index].scroll = content.scrollX + "," + content.scrollY;
				tabData.index = history.index + 1;
			}
			else if (history && history.count > 0)
			{
				for (var j = 0; j < history.count; j++)
				{
					tabData.Entry.push(this.serializeHistoryEntry(history.getEntryAtIndex(j, false), (j == history.index)?XPCNativeWrapper(browser.contentWindow):null));
				}
				tabData.index = history.index + 1;
				
				browser.__CR_data = tabData;
			}
			else
			{
				tabData.Entry[0] = { uri: browser.currentURI.spec };
			}
			tabData.zoom = (((browser.markupDocumentViewer.textZoom - 1) || -1) + 1) || "";
			
			var disallow = this.mCapabilities.filter(function(aCapability) {
				return !browser.docShell["allow" + aCapability];
			});
			tabData.disallow = disallow.join(",");
			
			var xulattr = Array.filter(tabbrowser.mTabs[i].attributes, function(aAttr) {
				return aAttr.name.substr(0, 3).toUpperCase() == "CR_" || CrashRecoveryService.mXulAttributes.indexOf(aAttr.name) > -1;
			}).map(function(aAttr) {
				return aAttr.name + "=" + encodeURI(aAttr.value);
			});
			tabData.xultab = xulattr.join(" ");
			
			tabs.push(tabData);
			
			if (browser == tabbrowser.selectedBrowser)
			{
				this.mWindows[aWindow.__CRi].selected = i + 1;
			}
		}
	},

	serializeHistoryEntry: function(aEntry, aContent)
	{
		var entry = { uri: aEntry.URI.spec, Child: [] };
		
		if (aEntry.title && aEntry.title != entry.uri)
		{
			entry.title = aEntry.title;
		}
		if (aEntry.isSubFrame)
		{
			entry.subframe = true;
		}
		if (!(aEntry instanceof Components.interfaces.nsISHEntry))
		{
			return entry;
		}
		
		if (aContent)
		{
			entry.scroll = aContent.scrollX + "," + aContent.scrollY;
		}
		else
		{
			var x = {}, y = {};
			aEntry.getScrollPosition(x, y);
			entry.scroll = x.value + "," + y.value;
		}
		entry.ID = aEntry.ID;
		
		try
		{
			if (this.mPref_postdata && aEntry.postData && this.checkPrivacyLevel(entry.uri))
			{
				if ("@mozilla.org/js_headerinfo;1" in Components.classes)
				{
					throw "LiveHTTPHeaders seems to crash Firefox at file uploads!";
				}
				
				aEntry.postData.QueryInterface(Components.interfaces.nsISeekableStream).seek(Components.interfaces.nsISeekableStream.NS_SEEK_SET, 0);
				var stream = this._COM("@mozilla.org/scriptableinputstream;1", "nsIScriptableInputStream", true);
				stream.init(aEntry.postData);
				var postdata = stream.read(stream.available());
				if (this.mPref_postdata == -1 || postdata.replace(/^(Content-.*\r\n)+(\r\n)*/, "").length <= this.mPref_postdata)
				{
					entry.postdata = postdata;
				}
				// should we close the stream here (since we never really opened it)?
			}
		}
		catch (ex) { debug(ex); }
		
		if (!(aEntry instanceof Components.interfaces.nsISHContainer))
		{
			return entry;
		}
		
		for (var i = 0; i < aEntry.childCount; i++)
		{
			var child = aEntry.GetChildAt(i);
			if (child)
			{
				entry.Child.push(this.serializeHistoryEntry(child, null));
			}
			else
			{
				entry.Child.push({ uri: "about:blank" });
			}
		}
		
		return entry;
	},

	saveTextData: function(aBrowser, aTextarea)
	{
		var id = (aTextarea.id)?"#" + aTextarea.id:aTextarea.name;
		if (!id || aTextarea.tagName.toLowerCase() != "textarea" && (aTextarea.tagName.toLowerCase() != "input" || aTextarea.type.toLowerCase() != "text"))
		{
			return;
		}
		if (!aBrowser.__CR_text)
		{
			aBrowser.__CR_text = {};
		}
		aBrowser.__CR_text[id] = aTextarea.value;
	},

	updateTextAndFramesetData: function(aWindow)
	{
		Array.forEach(aWindow.getBrowser().browsers, function(aBrowser, aIx) {
			try
			{
				var tabData = CrashRecoveryService.mWindows[aWindow.__CRi].Tab[aIx] || null;
				if (!tabData)
				{
					return;
				}
				
				var text = [];
				if (aBrowser.__CR_text && CrashRecoveryService.checkPrivacyLevel(aBrowser.currentURI.spec))
				{
					for (var id in aBrowser.__CR_text)
					{
						text.push(id + "=" + encodeURI(aBrowser.__CR_text[id]));
					}
				}
				if (aBrowser.currentURI.spec == "about:config")
				{
					text = ["#textbox=" + encodeURI(aBrowser.contentDocument.getElementById("textbox").value)];
				}
				tabData.text = text.join(" ");
				
				var framescroll = [];
				var document = XPCNativeWrapper(aBrowser.contentDocument);
				for (var i = 0; i < document.defaultView.frames.length; i++)
				{
					var frame = document.defaultView.frames[i];
					framescroll.push(frame.scrollX + "," + frame.scrollY);
				}
				tabData.framescroll = framescroll.join(" ");
			}
			catch (ex) { debug(ex); }
		});
	},

	updateCookieHosts: function(aWindow)
	{
		var hosts = this.mWindows[aWindow.__CRi]._hosts = [];
		
		function extractHosts(aEntry)
		{
			if (/^https?:\/\/(?:[^@\/\s]+@)?([\w.-]+)/.test(aEntry.uri))
			{
				var host = RegExp.$1;
				while (/^(?:[\w-]+\.([\w.-]+)|localhost)$/.test(host))
				{
					hosts.push(host);
					host = RegExp.$1;
				}
			}
			if (aEntry.Child)
			{
				aEntry.Child.forEach(extractHosts);
			}
		}
		
		this.mWindows[aWindow.__CRi].Tab.forEach(function(aTabData) { aTabData.Entry.forEach(extractHosts); });
	},

	getCookies: function(aHosts)
	{
		var cookiesEnum = this._COM("@mozilla.org/cookiemanager;1", "nsICookieManager").enumerator;
		var cookies = { count: 0 };
		
		while (cookiesEnum.hasMoreElements())
		{
			var cookie = cookiesEnum.getNext().QueryInterface(Components.interfaces.nsICookie);
			if (cookie.expires == 0 && cookie.host)
			{
				var host = cookie.host.replace(/^\./, "");
				if (aHosts.indexOf(host) > -1)
				{
					var url = "http" + ((cookie.isSecure)?"s":"") + "://" + cookie.host + (cookie.path || "").replace(/^(?!\/)/, "/");
					if (this.checkPrivacyLevel(url))
					{
						var value = (cookie.name || "name") + "=" + (cookie.value || "") + ";";
						value += (cookie.isDomain)?"domain=" + host + ";":"";
						value += (cookie.path)?"path=" + cookie.path + ";":"";
						value += (cookie.isSecure)?"secure;":"";
						
						cookies["domain" + ++cookies.count] = url;
						cookies["value" + cookies.count] = value;
					}
				}
			}
		}
		
		return (cookies.count > 0)?cookies:null;
	},

	updateWindowFeatures: function(aWindow)
	{
		var winData = this.mWindows[aWindow.__CRi];
		
		this.mWinAttributes.forEach(function(aAttr) {
			winData[aAttr] = CrashRecoveryService.getWindowDimension(aWindow, aAttr);
		});
		
		winData.hidden = this.mWinHidable.filter(function(aItem) {
			return aWindow[aItem] && !aWindow[aItem].visible;
		}).join(",");
		
		if (aWindow.toggleSidebar) // Firefox
		{
			winData.sidebar = aWindow.document.getElementById("sidebar-box").getAttribute("sidebarcommand");
		}
		else if (aWindow.sidebar_is_hidden) // SeaMonkey
		{
			winData.sidebar = !aWindow.sidebar_is_hidden();
		}
	},

	getCurrentState: function()
	{
		var activeWindow = this.getMostRecentBrowserWindow();
		
		if (this.mLoadState == 1)
		{
			this.getBrowserWindows().forEach(function(aWindow) {
				if (CrashRecoveryService.mDirty.all || CrashRecoveryService.mDirty[aWindow.__CRi] || aWindow == activeWindow)
				{
					CrashRecoveryService.saveWindowHistory(aWindow);
					CrashRecoveryService.updateTextAndFramesetData(aWindow);
					CrashRecoveryService.updateCookieHosts(aWindow);
				}
				CrashRecoveryService.updateWindowFeatures(aWindow);
			});
			this.mDirty = {};
		}
		
		var total = [], windows = [];
		for (var ix in this.mWindows)
		{
			if (this.mWindows[ix])
			{
				this.mWindows[ix].Cookies = this.getCookies(this.mWindows[ix]._hosts || []);
				total.push(this.mWindows[ix]);
				windows.push(ix);
			}
		}
		
		var ix = (activeWindow)?windows.indexOf(activeWindow.__CRi || ""):-1;
		if (ix > 0)
		{
			total.unshift(total.splice(ix, 1)[0]);
		}
		if (total.length == 0 && this.mLastWindowClosed)
		{
			total.push(this.mLastWindowClosed);
		}
		
		return IniObjectSerializer.encode({ Window: total });
	},

/* ........ Restoring Functionality .............. */

	restoreWindow_proxy: function(aWindow, aState, aOverwriteTabs)
	{
		CrashRecoveryService.restoreWindow(aWindow, aState, aOverwriteTabs);
	},

	restoreWindow: function(aWindow, aState, aOverwriteTabs)
	{
		try
		{
			var root = (typeof aState == "string")?IniObjectSerializer.decode(aState):aState;
			var winData = root.Window[0];
		}
		catch (ex)
		{
			debug(ex);
			return;
		}
		
		var opener = root.opener || null;
		if (!opener)
		{
			for (var w = 1; w < root.Window.length; w++)
			{
				this.openWindowWithState({ opener: aWindow, Window: [root.Window[w]] });
			}
		}
		
		if (!aWindow.getBrowser || !aWindow.getBrowser())
		{
			aWindow.setTimeout(this.restoreWindow_proxy, 100, aWindow, root, aOverwriteTabs);
			return;
		}
		
		if (aOverwriteTabs)
		{
			this.restoreWindowFeatures(aWindow, winData, opener);
		}
		if (winData.Cookies)
		{
			this.restoreCookies(winData.Cookies);
		}
		if (!winData.Tab)
		{
			winData.Tab = [];
		}
		
		var currentTabs = [];
		var tabbrowser = aWindow.getBrowser();
		var openTabCount = (aOverwriteTabs)?tabbrowser.browsers.length:-1;
		if (!tabbrowser.mTabs) // compatibility fix for MultiZilla
		{
			tabbrowser.mTabs = tabbrowser.mTabContainer.childNodes;
			tabbrowser.moveTabTo = tabbrowser.moveTabTo || tabbrowser.moveTab;
		}
		
		for (var t = 0; t < winData.Tab.length; t++)
		{
			currentTabs.push((t < openTabCount)?tabbrowser.mTabs[t]:tabbrowser.addTab());
			if (!aOverwriteTabs && root._firstTabs)
			{
				tabbrowser.moveTabTo(currentTabs[currentTabs.length - 1], t);
			}
		}
		for (t = openTabCount - 1; t >= currentTabs.length; t--)
		{
			tabbrowser.removeTab(tabbrowser.mTabs[t]);
		}
		
		aWindow.setTimeout(this.restoreHistory_proxy, 0, winData.Tab, currentTabs, (aOverwriteTabs)?winData.selected:0, aOverwriteTabs);
	},

	restoreHistory_proxy: function(aTabs, aCurrentTabs, aSelectTab, aCount)
	{
		CrashRecoveryService.restoreHistory(this, aTabs, aCurrentTabs, aSelectTab, aCount);
	},

	restoreHistory: function(aWindow, aTabs, aCurrentTabs, aSelectTab, aCount)
	{
		for (var t = 0; t < aTabs.length; t++)
		{
			var tab = aCurrentTabs[t];
			try
			{
				var browser = aWindow.getBrowser().getBrowserForTab(tab);
				var history = browser.webNavigation.sessionHistory;
				
				if (!history)
				{
					throw new Error();
				}
			}
			catch (ex)
			{
				aCount = (aCount || 0) + 1;
				if (aCount < 10)
				{
					aWindow.setTimeout(this.restoreHistory_proxy, 100, aTabs, aCurrentTabs, aSelectTab, aCount);
					return;
				}
				continue;
			}
			
			if (history.count > 0)
			{
				history.PurgeHistory(history.count);
			}
			history.QueryInterface(Components.interfaces.nsISHistoryInternal);
			
			var tabData = aTabs[t];
			var activeIndex = (tabData.index || 0) - 1;
			browser.markupDocumentViewer.textZoom = parseFloat(tabData.zoom || 1);
			
			var currentScrollPos = [0, 0];
			
			for (var i = 0; i < tabData.Entry.length; i++)
			{
				history.addEntry(this.deserializeHistoryEntry(tabData.Entry[i]), true);
				
				if (i == activeIndex)
				{
					var pos = (tabData.Entry[i].scroll || "0,0").split(",");
					currentScrollPos = [parseInt(pos[0]) || 0, parseInt(pos[1] || "") || 0];
				}
			}
			if (tabData.disallow)
			{
				tabData.disallow.split(",").forEach(function(aCapability) {
					browser.docShell["allow" + aCapability] = false;
				});
			}
			if (tabData.xultab)
			{
				tabData.xultab.split(" ").forEach(function(aAttr) {
					if (/^([^\s=]+)=(.*)/.test(aAttr))
					{
						tab.setAttribute(RegExp.$1, decodeURI(RegExp.$2));
					}
				});
			}
			
			if (t == aSelectTab - 1)
			{
				aWindow.getBrowser().selectedTab = tab;
			}
			
			var event = aWindow.document.createEvent("Events");
			event.initEvent("CRTabRestoring", true, false);
			tab.dispatchEvent(event);
			
			try
			{
				browser.webNavigation.gotoIndex(activeIndex);
			}
			catch (ex) { }
			
			var id = "__CR" + t + "_" + Date.now();
			browser[id] = eval(this.restoreDocument_proxy.toString()
				.replace(/__ID__/g, id)
				.replace(/__SCROLL_POSITION__/g, currentScrollPos.join(", "))
				.replace(/__TEXT_ARRAY__/g, (tabData.text)?"\"" + tabData.text.replace(/[\\"]/g, '\\$&') + "\".split(' ')":"[]")
				.replace(/__FRAMESCROLL_ARRAY__/g, (tabData.framescroll)?"\"" + tabData.framescroll.replace(/[\\"]/g, '\\$&') + "\".split(' ')":"[]"));
			browser[id + "_tab"] = tab;
			browser.addEventListener("load", browser[id], true);
		}
	},

	deserializeHistoryEntry: function(aEntry)
	{
		var shEntry = this._COM("@mozilla.org/browser/session-history-entry;1", "nsISHEntry", true);
		if (aEntry.ID)
		{
			shEntry.ID = aEntry.ID;
		}
		shEntry.setURI(this.mIOService.newURI(aEntry.uri, null, null));
		shEntry.setTitle(aEntry.title || aEntry.uri);
		shEntry.setIsSubFrame(aEntry.subframe || false);
		shEntry.loadType = Components.interfaces.nsIDocShellLoadInfo.loadHistory;
		
		var scrollPos = (aEntry.scroll || "0,0").split(",");
		scrollPos = [parseInt(scrollPos[0]) || 0, parseInt(scrollPos[1]) || 0];
		shEntry.setScrollPosition(scrollPos[0], scrollPos[1]);
		
		if (aEntry.postdata)
		{
			var stream = this._COM("@mozilla.org/io/string-input-stream;1", "nsIStringInputStream", true);
			stream.setData(aEntry.postdata, -1);
			shEntry.postData = stream;
		}
		
		if (aEntry.Child && shEntry instanceof Components.interfaces.nsISHContainer)
		{
			for (var i = 0; i < aEntry.Child.length; i++)
			{
				shEntry.AddChild(this.deserializeHistoryEntry(aEntry.Child[i]), i);
			}
		}
		
		return shEntry;
	},

	restoreDocument_proxy: function(aEvent)
	{
		if (aEvent && aEvent.originalTarget && aEvent.originalTarget.defaultView && aEvent.originalTarget.defaultView == aEvent.originalTarget.defaultView.top)
		{
			function restoreTextData(aContent)
			{
				__TEXT_ARRAY__.forEach(function(aEntry) {
					if (/^(#?)([\w-]+)=(.*)$/.test(aEntry))
					{
						var document = aContent.document;
						var node = (RegExp.$1)?document.getElementById(RegExp.$2):document.getElementsByName(RegExp.$2)[0] || null;
						if (node && "value" in node)
						{
							node.value = decodeURI(RegExp.$3);
							
							var event = document.createEvent("UIEvents");
							event.initUIEvent("input", true, true, aContent, 0);
							node.dispatchEvent(event);
						}
					}
				});
			}
			
			var content = XPCNativeWrapper(aEvent.originalTarget).defaultView;
			if (this.currentURI.spec == "about:config")
			{
				content = aEvent.originalTarget.defaultView;
			}
			content.scrollTo(__SCROLL_POSITION__);
			restoreTextData(content);
			
			__FRAMESCROLL_ARRAY__.forEach(function(aEntry, aIx) {
				if (content.frames[aIx])
				{
					if (/^(\d+),(\d+)$/.test(aEntry))
					{
						content.frames[aIx].scrollTo(RegExp.$1, RegExp.$2);
					}
					restoreTextData(content.frames[aIx]);
				}
			});
			
			var event = aWindow.document.createEvent("Events");
			event.initEvent("CRTabRestored", true, false);
			this.__ID___tab.dispatchEvent(event);
			
			this.removeEventListener("load", this.__ID__, true);
			delete this.__ID___tab;
			delete this.__ID__;
		}
	},

	restoreWindowFeatures: function(aWindow, aWinData, aOpener)
	{
		if (aWinData.hidden)
		{
			aWinData.hidden.split(",").forEach(function(aDecor) {
				if (aWindow[aDecor])
				{
					aWindow[aDecor].visible = false;
				}
			});
		}
		
		aWindow.setTimeout(this.restoreDimensions_proxy, 0, aOpener, aWinData.width, aWinData.height, aWinData.screenX, aWinData.screenY, aWinData.sizemode, aWinData.sidebar || "");
	},

	restoreDimensions_proxy: function(aOpener, aWidth, aHeight, aLeft, aTop, aSizeMode, aSidebar)
	{
		var win = this;
		function win_(aName) { return CrashRecoveryService.getWindowDimension(win, aName); }
		
		if (aWidth && aHeight && (aWidth != win_("width") || aHeight != win_("height")))
		{
			this.resizeTo(aWidth, aHeight);
		}
		if (!isNaN(aLeft) && !isNaN(aTop) && (aLeft != win_("screenX") || aTop != win_("screenY")))
		{
			this.moveTo(aLeft, aTop);
		}
		if (aSizeMode == "maximized" && win_("sizemode") != "maximized")
		{
			this.maximize();
		}
		else if (aSizeMode && aSizeMode != "maximized" && win_("sizemode") != "normal")
		{
			this.restore();
		}
		var sidebar = this.document.getElementById("sidebar-box");
		if (this.toggleSidebar && sidebar.getAttribute("sidebarcommand") != aSidebar)
		{
			this.toggleSidebar(aSidebar);
		}
		else if (this.sidebar_is_hidden && this.sidebar_is_hidden() != !aSidebar)
		{
			this.SidebarShowHide();
		}
		if (aOpener)
		{
			aOpener.focus();
		}
	},

	restoreCookies: function(aCookies)
	{
		var cookieService = this._COM("@mozilla.org/cookieService;1", "nsICookieService");
		
		for (var i = 1; i <= aCookies.count; i++)
		{
			try
			{
				cookieService.setCookieString(this.mIOService.newURI(aCookies["domain" + i], null, null), null, aCookies["value" + i] + "expires=0;", null);
			}
			catch (ex) { debug(ex); }
		}
	},

	// code adapted from Danil Ivanov's "Cache Fixer" extension
	restoreCache: function()
	{
		var cache = this._COM("@mozilla.org/file/directory_service;1", "nsIProperties").get("ProfLD", Components.interfaces.nsILocalFile);
		cache.append("Cache");
		cache.append("_CACHE_MAP_");
		if (!cache.exists())
		{
			return;
		}
		
		var stream = this._COM("@mozilla.org/network/file-input-stream;1", "nsIFileInputStream", true);
		stream.init(cache, 0x01, 0, 0); // PR_RDONLY
		var input = this._COM("@mozilla.org/binaryinputstream;1", "nsIBinaryInputStream", true);
		input.setInputStream(stream);
		var content = input.readByteArray(input.available());
		input.close();
		stream.close();
		
		if (content[15] != 1)
		{
			return;
		}
		content[15] = 0;
		
		stream = this._COM("@mozilla.org/network/file-output-stream;1", "nsIFileOutputStream", true);
		stream.init(cache, 0x02 | 0x20, 0600, 0); // PR_WRONLY | PR_TRUNCATE
		var output = this._COM("@mozilla.org/binaryoutputstream;1", "nsIBinaryOutputStream", true);
		output.setOutputStream(stream);
		output.writeByteArray(content, content.length);
		output.flush();
		output.close();
		stream.close();
	},

	retryDownloads: function(aWindow)
	{
		var _COM = CrashRecoveryService._COM;
		
		function AutoDownloader(aUrl, aFilename)
		{
			this.url = aUrl;
			this.filename = aFilename;
		}
		AutoDownloader.prototype = {
			onStartRequest: function(aRequest, aContext) { },
			onStopRequest: function(aRequest, aContext, aStatus)
			{
				if (aStatus == Components.results.NS_OK)
				{
					var file = _COM("@mozilla.org/file/local;1", "nsILocalFile", true);
					file.initWithPath(this.filename);
					if (file.exists())
					{
						file.remove(false);
					}
					
					aWindow.saveURL(this.url, this.filename, null, true, true, null);
				}
			}
		};
		
		var downloadManager = _COM("@mozilla.org/download-manager;1", "nsIDownloadManager");
		var rdfService = _COM("@mozilla.org/rdf/rdf-service;1", "nsIRDFService");
		var ioService = _COM("@mozilla.org/network/io-service;1", "nsIIOService");
		
		var rdfContainer = _COM("@mozilla.org/rdf/container;1", "nsIRDFContainer", true);
		var datasource = downloadManager.datasource;
		
		try
		{
			rdfContainer.Init(datasource, rdfService.GetResource("NC:DownloadsRoot"));
		}
		catch (ex) // SeaMonkey doesn't support saveURL without user interaction, so we stop here (it doesn't support downloadManager.datasource, either)
		{
			return;
		}
		
		var downloads = rdfContainer.GetElements();
		
		while (downloads.hasMoreElements())
		{
			var download = downloads.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
			
			var node = datasource.GetTarget(rdfService.GetResource(download.Value), rdfService.GetResource("http://home.netscape.com/NC-rdf#DownloadState"), true);
			if (node)
			{
				node.QueryInterface(Components.interfaces.nsIRDFInt);
			}
			if (!node || node.Value != Components.interfaces.nsIDownloadManager.DOWNLOAD_DOWNLOADING)
			{
				continue;
			}
			
			node = datasource.GetTarget(rdfService.GetResource(download.Value), rdfService.GetResource("http://home.netscape.com/NC-rdf#URL"), true);
			node.QueryInterface(Components.interfaces.nsIRDFResource);
			
			var url = node.Value;
			
			node = datasource.GetTarget(rdfService.GetResource(download.Value), rdfService.GetResource("http://home.netscape.com/NC-rdf#File"), true);
			node.QueryInterface(Components.interfaces.nsIRDFResource);
			
			var linkChecker = _COM("@mozilla.org/network/urichecker;1", "nsIURIChecker", true);
			
			linkChecker.init(ioService.newURI(url, null, null));
			linkChecker.loadFlags = Components.interfaces.nsIRequest.LOAD_BACKGROUND;
			linkChecker.asyncCheck(new AutoDownloader(url, node.Value), null);
		}
	},

/* ........ Disk Access .............. */

	saveStateDelayed_proxy: function()
	{
		CrashRecoveryService.mSavePending = false;
		CrashRecoveryService.saveStateDelayed(this, -1);
	},

	saveStateDelayed: function(aWindow, aDelay)
	{
		var minimalDelay = this.mLastSaveTime + this.mPref_interval - Date.now();
		if (this.mSavePending && minimalDelay < -this.mPref_interval)
		{
			this.mSavePending = false;
			this.mLastSaveTime = this.mLastSaveTime - minimalDelay;
		}
		
		this.mDirty[aWindow.__CRi] = true;
		if (!this.mSavePending)
		{
			aDelay = Math.max(minimalDelay, aDelay || 2000);
			if (aDelay > 0)
			{
				this.mSavePending = true;
				aWindow.setTimeout(this.saveStateDelayed_proxy, aDelay);
			}
			else
			{
				this.saveState();
				this.mLastSaveTime = Date.now();
			}
		}
	},

	saveState: function(aUpdateAll)
	{
		this.mDirty.all = aUpdateAll;
		this.stateToDisk([
			"; This file will be periodically overwritten by Crash Recovery",
			"; Do not edit while Firefox is running",
			"; The file encoding is UTF-8",
			"[CrashRecovery]",
			"state=" + ((this.mLoadState > 0)?"running":"stopped"),
			"date=" + (new Date()).toString(),
			this.getCurrentState(),
			""
		].join("\n").replace(/\n\[/g, "\n$&"), aUpdateAll);
	},

	stateToDisk: function(aState, aThisThread, aBackup)
	{
		this.writeFile(this.getSessionFile(aBackup), aState, aThisThread);
	},

	stateFromDisk: function()
	{
		return this.readFile(this.getSessionFile()) || "";
	},

	clearDisk: function()
	{
		this.delFile(this.getSessionFile());
		this.delFile(this.getSessionFile(true));
	},

	getSessionFile: function(aBackup)
	{
		var file = this.mProfileDirectory.clone();
		
		file.append("crashrecovery." + ((aBackup)?"bak":"dat"));
		
		return file;
	},

/* ........ Auxiliary Functions .............. */

	getBrowserWindows: function()
	{
		var windowsEnum = this.mWindowMediator.getEnumerator("navigator:browser");
		var windows = [];
		
		while (windowsEnum.hasMoreElements())
		{
			var window = windowsEnum.getNext();
			if (window.__CRi)
			{
				windows.push(window);
			}
		}
		
		return windows;
	},

	getMostRecentBrowserWindow: function()
	{
		return this.mWindowMediator.getMostRecentWindow("navigator:browser");
	},

	openWindowWithState: function(aState)
	{
		var argString = this._COM("@mozilla.org/supports-string;1", "nsISupportsString", true);
		argString.data = "";
		
		// Why isn't it possible to set the window's dimensions here (as feature)?
		var window = this._COM("@mozilla.org/embedcomp/window-watcher;1", "nsIWindowWatcher").openWindow(null, this.getPref("browser.chromeURL"), "_blank", "chrome,dialog=no,all", argString);
		
		window.__CR_state = aState;
		window.addEventListener("load", this.openWindowWithState_onLoad_proxy, true);
	},

	openWindowWithState_onLoad_proxy: function()
	{
		this.removeEventListener("load", CrashRecoveryService.openWindowWithState_onLoad_proxy, true);
		CrashRecoveryService.restoreWindow(this, this.__CR_state, true);
		delete this.__CR_state;
	},

	doResumeSession: function()
	{
		return this.mPref_resume_session || this.getPref("browser.startup.page", 1) == 2;
	},

	doRecoverSession: function()
	{
		if (!this.getPref(this.mPrefRoot + "restore_prompt", true))
		{
			return true;
		}
		
		var recover = true;
		var dontPrompt = { value: false };
		var dialogURI = this.getPref(this.mPrefRoot + "restore_prompt_uri");
		
		try
		{
			if (dialogURI)
			{
				var params = this._COM("@mozilla.org/embedcomp/dialogparam;1", "nsIDialogParamBlock", true);
				params.SetInt(0, 0);
				params.SetInt(1, 0);
				this._COM("@mozilla.org/embedcomp/window-watcher;1", "nsIWindowWatcher").openWindow(null, dialogURI, "_blank", "chrome,modal,centerscreen,titlebar", params);
				recover = params.GetInt(0) == 0;
				dontPrompt.value = params.GetInt(1);
			}
			else
			{
				recover = this._COM("@mozilla.org/embedcomp/prompt-service;1", "nsIPromptService").confirmCheck(null, "Crash Recovery", this.mStringBundle.GetStringFromName("restore_prompt"), this.mStringBundle.GetStringFromName("restore_prompt_again"), dontPrompt);
			}
		}
		catch (ex) { }
		
		if (dontPrompt.value)
		{
			this.setPref(this.mPrefRoot + "restore_prompt", false);
		}
		
		return recover;
	},

	isCmdLineEmpty: function(aWindow)
	{
		if (!aWindow.arguments)
		{
			return true;
		}
		
		var homepage = null;
		switch (this.getPref("browser.startup.page", 1))
		{
		case 0:
			homepage = "about:blank";
			break;
		case 1:
			try
			{
				homepage = this.mPrefBranch.getComplexValue("browser.startup.homepage", Components.interfaces.nsIPrefLocalizedString).data;
			}
			catch (ex)
			{
				homepage = this.getPref("browser.startup.homepage", "");
			}
			break;
		case 2:
			homepage = this._COM("@mozilla.org/browser/global-history;2", "nsIBrowserHistory").lastPageVisited;
			break;
		}
		
		for (var i = 0; i < aWindow.arguments.length; i++)
		{
			var url = aWindow.arguments[i].split("\n")[0];
			if (!url || url == homepage)
			{
				aWindow.arguments.splice(i--, 1);
			}
		}
		
		return (aWindow.arguments.length == 0);
	},

	checkPrivacyLevel: function(aURL)
	{
		return this.mPref_privacy_level < ((aURL.substr(0, 6) == "https:")?1:2);
	},

	getWindowDimension: function(aWindow, aAttribute)
	{
		if (aAttribute == "sizemode")
		{
			return (aWindow.windowState == aWindow.STATE_MAXIMIZED)?"maximized":"normal";
		}
		
		var equivalents = { width: "outerWidth", height: "outerHeight" };
		var attribute = equivalents[aAttribute] || aAttribute;
		var dimension = (attribute in aWindow)?aWindow[attribute]:"";
		
		if (aWindow.windowState == aWindow.STATE_NORMAL)
		{
			return dimension;
		}
		return aWindow.document.documentElement.getAttribute(aAttribute) || dimension;
	},

	ensureBrowser: function(aBrowser)
	{
		return (aBrowser.localName != "browser")?aBrowser.childNodes[1]:aBrowser;
	},

	getStringBundle: function(aURI)
	{
		return this._COM("@mozilla.org/intl/stringbundle;1", "nsIStringBundleService").createBundle(aURI, this._COM("@mozilla.org/intl/nslocaleservice;1", "nsILocaleService").getApplicationLocale());
	},

	_COM: function(aClassName, aInterface, aNew)
	{
		return Components.classes[aClassName][(aNew)?"createInstance":"getService"](Components.interfaces[aInterface]);
	},

/* ........ Storage API .............. */

	getPref: function(aName, aDefault)
	{
		var pb = this.mPrefBranch;
		switch (pb.getPrefType(aName))
		{
		case pb.PREF_STRING:
			return pb.getCharPref(aName);
		case pb.PREF_BOOL:
			return pb.getBoolPref(aName);
		case pb.PREF_INT:
			return pb.getIntPref(aName);
		default:
			return aDefault;
		}
	},

	setPref: function(aName, aValue)
	{
		switch (typeof(aValue))
		{
		case "boolean":
			this.mPrefBranch.setBoolPref(aName, aValue);
			break;
		case "number":
			this.mPrefBranch.setIntPref(aName, parseInt(aValue));
			break;
		default:
			this.mPrefBranch.setCharPref(aName, "" + aValue);
			break;
		}
	},

	readFile: function(aFile)
	{
		try
		{
			var stream = this._COM("@mozilla.org/network/file-input-stream;1", "nsIFileInputStream", true);
			stream.init(aFile, 0x01, 0, 0);
			var sstream = this._COM("@mozilla.org/scriptableinputstream;1", "nsIScriptableInputStream", true);
			sstream.init(stream);
			
			var data = sstream.read(-1).replace(/\r?\n/g, "\n");
			sstream.close();
			stream.close();
			
			return this.mUnicodeConverter.ConvertToUnicode(data);
		}
		catch (ex) { }
		
		return null;
	},

	writeFile: function(aFile, aData, aThisThread)
	{
		if (aThisThread)
		{
			(new this.FileWriter(aFile, aData)).run();
			return;
		}
		
		var nsIThread = Components.interfaces.nsIThread;
		this._COM("@mozilla.org/thread;1", "nsIThread", true).init(
			new this.FileWriter(aFile, aData),
			128 * 1024,
			nsIThread.PRIORITY_NORMAL,
			nsIThread.SCOPE_GLOBAL,
			nsIThread.STATE_UNJOINABLE
		);
	},

	initFileWriter: function()
	{
		this.FileWriter_prototype.mEOL = this.getEOL();
		this.FileWriter_prototype.mUnicodeConverter = this.mUnicodeConverter;
		this.FileWriter.prototype = this.FileWriter_prototype;
	},

	FileWriter: function(aFile, aData)
	{
		this.mFile = aFile;
		this.mData = aData;
	},

	FileWriter_prototype: {
		run: function()
		{
			if (!this.mFile.exists())
			{
				this.mFile.create(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0600);
			}
			var stream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
			var data = this.mUnicodeConverter.ConvertFromUnicode(this.mData).replace(/\n/g, this.mEOL);
			
			stream.init(this.mFile, 0x02 | 0x08 | 0x20, 0600, 0);
			stream.write(data, data.length);
			stream.flush();
			stream.close();
		}
	},

	delFile: function(aFile)
	{
		try
		{
			aFile.remove(false);
		}
		catch (ex) { }
	},

	getEOL: function()
	{
		if (!Components.interfaces.nsIXULRuntime) // SeaMonkey isn't based on XULRunner (yet)
		{
			return "\r\n";
		}
		
		var OS = this._COM("@mozilla.org/xre/app-info;1", "nsIXULRuntime").OS;
		
		return /win|os[\/_]?2/i.test(OS)?"\r\n":/mac/i.test(OS)?"\r":"\n";
	},

/* ........ QueryInterface .............. */

	QueryInterface: function(aIID)
	{
		if (!aIID.equals(Components.interfaces.nsISupports) && !aIID.equals(Components.interfaces.nsIObserver) && !aIID.equals(this.mCID))
		{
			Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
			return null;
		}
		
		return this;
	}
};

/* :::::::: File Format Converter ::::::::::::::: */

/*const*/ var IniObjectSerializer = {
	encode: function(aObj, aPrefix)
	{
		aPrefix = (aPrefix)?aPrefix + ".":"";
		
		var ini = [], iniChildren = [];
		
		for (var key in aObj)
		{
			if (!key || key.charAt(0) == ";" || key.charAt(0) == "[" || key.indexOf("]") > -1 || key.indexOf("=") > -1)
			{
				debug("Ignoring invalid key name: '" + key + "'!");
				continue;
			}
			else if (key.charAt(0) == "_")
			{
				continue;
			}
			
			var value = aObj[key];
			if (typeof value == "boolean" || typeof value == "number")
			{
				ini.push(key + "=" + value);
			}
			else if (typeof value == "string" && value)
			{
				ini.push(key + "=" + ((/[%\t\r\n]/.test(value))?encodeURI(value):value));
			}
			else if (value instanceof Array)
			{
				for (var i = 0; i < value.length; i++)
				{
					if (value[i] instanceof Object)
					{
						iniChildren.push("[" + aPrefix + key + (i + 1) + "]");
						iniChildren.push(this.encode(value[i], aPrefix + key + (i + 1)));
					}
				}
			}
			else if (typeof value == "object" && value)
			{
				iniChildren.push("[" + aPrefix + key + "]");
				iniChildren.push(this.encode(value, aPrefix + key));
			}
		}
		
		return ini.concat(iniChildren).join("\n");
	},

	decode: function(aIniString)
	{
		if (this.CR_isOldFormat(aIniString))
		{
			return this.CR_decodeOldFormat(aIniString);
		}
		
		var rootObject = {};
		var obj = rootObject;
		var lines = aIniString.split("\n");
		
		for (var i = 0; i < lines.length; i++)
		{
			try
			{
				if (lines[i].charAt(0) == "[")
				{
					obj = this._getObjForHeader(rootObject, lines[i]);
				}
				else if (lines[i] && lines[i].charAt(0) != ";")
				{
					this._setValueForLine(obj, lines[i]);
				}
			}
			catch (ex)
			{
				throw new Error("Error at line " + (i + 1) + ": " + ex.description);
			}
		}
		
		return rootObject;
	},

	_getObjForHeader: function(aObj, aLine)
	{
		var names = aLine.split("]")[0].substr(1).split(".");
		
		for (var i = 0; i < names.length; i++)
		{
			var name = names[i];
			if (!names[i])
			{
				throw new Error("Invalid header: [" + names.join(".") + "]!");
			}
			if (/(\d+)$/.test(names[i]))
			{
				names[i] = names[i].slice(0, -RegExp.$1.length);
				var ix = parseInt(RegExp.$1) - 1;
				aObj = aObj[names[i]] = aObj[names[i]] || [];
				aObj = aObj[ix] = aObj[ix] || {};
			}
			else
			{
				aObj = aObj[names[i]] = aObj[names[i]] || [];
			}
		}
		
		return aObj;
	},

	_setValueForLine: function(aObj, aLine)
	{
		var ix = aLine.indexOf("=");
		if (ix < 1)
		{
			throw new Error("Invalid entry: " + aLine + "!");
		}
		
		var value = aLine.substr(ix + 1);
		if (value == "true" || value == "false")
		{
			value = (value == "true");
		}
		else if (/^\d+$/.test(value))
		{
			value = parseInt(value);
		}
		else if (value.indexOf("%") > -1)
		{
			value = decodeURI(value);
		}
		
		aObj[aLine.substr(0, ix)] = value;
	},

	CR_isOldFormat: function(aString)
	{
		return /^.*~~~~/.test(aString);
	},

	CR_decodeOldFormat: function(aString)
	{
		aString = aString.replace(/^(.*?) ~~~~( .*)?/, "$1");
		var root = { CrashRecovery: { state: (RegExp.$2)?"running":"stopped" } };
		
		root.Window = aString.split("\n- WINDOW -----\n").map(function(aWindow) {
			var winData = {};
			var tabs = aWindow.split("\n-- TAB ----\n");
			var cookies = tabs.shift().split("\n");
			var features = cookies.shift();
			
			winData.selected = parseInt(features) + 1;
			if (/ dims:(\S+)/.test(features))
			{
				RegExp.$1.split(",").forEach(function(aFeat) {
					if (/^(\S+)=(.*)/.test(aFeat)) winData[RegExp.$1] = decodeURI(RegExp.$2);
				});
			}
			if (/ hide:(\S+)/.test(features)) winData.hidden = RegExp.$1;
			winData.sidebar = (/ sidebar:(\S+)/.test(features))?RegExp.$1:"";
			
			cookies.forEach(function(aCookie) {
				if (!window.Cookies) window.Cookies = { count: 0 };
				var cookie = aCookie.split(" ");
				window.Cookies["domain" + ++window.Cookies.count] = cookie[0];
				window.Cookies["value" + window.Cookies.count] = decodeURI(cookie[1]);
			});
			
			winData.Tab = tabs.map(IniObjectSerializer.CR_decodeOldFormatTab);
			
			return winData;
		});
		
		return root;
	},

	CR_decodeOldFormatTab: function(aTab)
	{
		function parseEntry(aTabs, aIx, aPrefix)
		{
			aPrefix = aPrefix || "";
			var entries = aTabs[aIx].split(" ");
			var entry = {
				uri: entries[0].substr(aPrefix.length), scroll: entries[1],
				postdata: decodeURI(entries[2] || ""), title: decodeURI(entries[3] || ""),
				subframe: (entries[4] || "") == "1", Child: []
			};
			
			aPrefix += "+";
			while (aTabs[aIx + 1] && aTabs[aIx + 1].substr(0, aPrefix.length) == aPrefix)
			{
				entry.Child.push(parseEntry(aTabs, aIx + 1, aPrefix));
				aTabs.splice(aIx + 1, 1);
			}
			
			return entry;
		}
		
		var tabData = { Entry: [], index: 0 };
		var tabs = aTab.split("\n");
		if (/^(\d+) zoom:(\S+)/.test(tabs.shift()))
		{
			tabData.index = parseInt(RegExp.$1) + 1;
			tabData.zoom = RegExp.$2;
		}
		
		for (var i = 0; i < tabs.length; i++)
		{
			if (tabs[i].charAt(0) == "@") break;
			tabData.Entry.push(parseEntry(tabs, i));
		}
		for (; i < tabs.length; i++)
		{
			if (/^@(text|framescroll|xulattr|disallow) (.*)/.test(tabs[i]))
			{
				tabData[RegExp.$1] = RegExp.$2;
			}
		}
		
		return tabData;
	}
};

/* :::::::: Miscellanea ::::::::::::::: */

function debug(aMsg)
{
	Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService).logStringMessage(("Crash Recovery: " + aMsg).replace(/\S{80}/g, "$&\n"));
}

/* :::::::: Service Registration & Initialization ::::::::::::::: */

/*const*/ var CrashRecoveryModuleAndFactory = {

/* ........ nsIModule .............. */

	getClassObject: function(aCompMgr, aCID, aIID)
	{
		if (aCID.equals(CrashRecoveryService.mCID))
		{
			return this.QueryInterface(aIID);
		}
		
		Components.returnCode = Components.results.NS_ERROR_NOT_REGISTERED;
		return null;
	},

	registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
	{
		aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
		aCompMgr.registerFactoryLocation(CrashRecoveryService.mCID, CrashRecoveryService.mClassName, CrashRecoveryService.mContractID, aFileSpec, aLocation, aType);
		
		var catMan = Components.classes["@mozilla.org/categorymanager;1"].getService(Components.interfaces.nsICategoryManager);
		catMan.addCategoryEntry("app-startup", CrashRecoveryService.mClassName, "service," + CrashRecoveryService.mContractID, true, true, null);
	},

	unregisterSelf: function(aCompMgr, aLocation, aType)
	{
		aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
		aCompMgr.unregisterFactoryLocation(CrashRecoveryService.mCID, aLocation);
		
		var catMan = Components.classes["@mozilla.org/categorymanager;1"].getService(Components.interfaces.nsICategoryManager);
		catMan.deleteCategoryEntry("app-startup", "service," + CrashRecoveryService.mContractID, true);
	},

	canUnload: function(aCompMgr)
	{
		return true;
	},

/* ........ nsIFactory .............. */

	createInstance: function(aOuter, aIID)
	{
		if (aOuter != null)
		{
			Components.returnCode = Components.results.NS_ERROR_NO_AGGREGATION;
			return null;
		}
		
		return CrashRecoveryService.QueryInterface(aIID);
	},

	lockFactory: function(aLock) { },

/* ........ QueryInterface .............. */

	QueryInterface: function(aIID)
	{
		if (!aIID.equals(Components.interfaces.nsISupports) && !aIID.equals(Components.interfaces.nsIModule) && !aIID.equals(Components.interfaces.nsIFactory))
		{
			Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
			return null;
		}
		
		return this;
	}
};

function NSGetModule(aComMgr, aFileSpec)
{
	return CrashRecoveryModuleAndFactory;
}
