/*const*/ var gSessionManager = {
	mCrashRecoveryService: Components.classes["@zeniko/crashrecoveryservice;1"].getService(Components.interfaces.nsICrashRecoveryService),
	mObserverService: Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService),
	mPrefBranch: Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch2),
	mWindowMediator: Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator),
	mUnicodeConverter: Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter),
	mPromptService: Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService),
	mProfileDirectory: Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties).get("ProfD", Components.interfaces.nsILocalFile),
	mIOService: Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService),
	mComponents: Components,

	mObserving: ["crashrecovery:windowclosed", "browser:purge-session-history", "quit-application-granted", "quit-application"],
	mPrefRoot: "extensions.sessionmanager.",

	mOneTabRE: /\n\[Window(\d+\.Tab\d+)\]\n[\s\S]*?(?=\n\[Window(?!\1)|$)/g,
	mClosedWindowFile: "sessionmanager.dat",
	mSessionExt: ".session",
	mPromptSessionName: "?",

	mClosedTabs: [],

/* ........ Listeners / Observers.............. */

	onLoad_proxy: function()
	{
		this.removeEventListener("load", gSessionManager.onLoad_proxy, false);
		gSessionManager.onLoad();
	},

	onLoad: function(aDialog)
	{
		this.mBundle = document.getElementById("bundle_sessionmanager");
		this.mTitle = this._string("sessionManager");
		this.mUnicodeConverter.charset = "UTF-8";
		this.mEOL = this.getEOL();
		
		if (aDialog || this.mFullyLoaded)
		{
			return;
		}
		
		this.mObserving.forEach(function(aTopic) {
			gSessionManager.mObserverService.addObserver(gSessionManager, aTopic, false);
		});
		
		this.mPref_backup_session = this.getPref(this.mPrefRoot + "backup_session", 1);
		this.mPref_max_closed_undo = this.getPref(this.mPrefRoot + "max_closed_undo", 10);
		this.mPref_max_tabs_undo = this.getPref(this.mPrefRoot + "max_tabs_undo", 10);
		this.mPref_name_format = this.getPref(this.mPrefRoot + "name_format", "%40t-%d");
		this.mPref_overwrite = this.getPref(this.mPrefRoot + "overwrite", false);
		this.mPref_reload = this.getPref(this.mPrefRoot + "reload", false);
		this.mPref_resume_session = this.getPref(this.mPrefRoot + "resume_session", "");
		this.mPref_submenus = this.getPref(this.mPrefRoot + "submenus", false);
		this.mPref__running = this.getPref(this.mPrefRoot + "_running", false);
		this.mPrefBranch.addObserver(this.mPrefRoot, this, false);
		
		getBrowser().addEventListener("CRTabRestored", this.onTabRestored_proxy, false);
		
		this.recoverSession(this.mPref__running);
		this.patchTabBrowser();
		this.updateToolbarButton();
		
		this.setPref(this.mPrefRoot + "_running", true);
		this.mFullyLoaded = true;
	},

	onUnload_proxy: function()
	{
		this.removeEventListener("unload", gSessionManager.onUnload_proxy, false);
		gSessionManager.onUnload();
	},

	onUnload: function()
	{
		this.mObserving.forEach(function(aTopic) {
			gSessionManager.mObserverService.removeObserver(gSessionManager, aTopic);
		});
		this.mPrefBranch.removeObserver(this.mPrefRoot, this);
		
		gBrowser.removeEventListener("CRTabRestored", this.onTabRestored_proxy, false);
		
		if (this.mPref__running && this.getBrowserWindows().length == 0)
		{
			this.mObserverService.addObserver(this, "quit-application", false);
			this.mObserverService.addObserver(this, "domwindowopened", false);
			
			this._string_preserve_session = this._string("preserve_session");
			this._string_backup_session = this._string("backup_session");
		}
		
		this.mBundle = null;
		this.mSessions = null;
		this.mClosedTabs = null;
		this.mFullyLoaded = false;
	},

	observe: function(aSubject, aTopic, aData)
	{
		switch (aTopic)
		{
		case "crashrecovery:windowclosed":
			if (aSubject == window)
			{
				this.appendClosedWindow(aData, content.document.title);
			}
			else
			{
				this.updateToolbarButton(true);
			}
			break;
		case "browser:purge-session-history":
			this.clearUndoData("all");
			this.delFile(this.getSessionDir("backup" + this.mSessionExt));
			break;
		case "nsPref:changed":
			var len = this.mPrefRoot.length;
			if (aData.substr(0, len) == this.mPrefRoot)
			{
				this["mPref_" + aData.substr(len)] = this.getPref(aData);
				
				switch (aData.substr(len))
				{
				case "max_closed_undo":
					if (this.mPref_max_closed_undo == 0)
					{
						this.clearUndoData("window", true);
					}
					break;
				case "max_tabs_undo":
					if (this.mPref_max_tabs_undo == 0)
					{
						this.clearUndoData("tab", true);
					}
					else
					{
						this.patchTabBrowser();
					}
					break;
				}
			}
			break;
		case "quit-application-granted":
		case "quit-application":
			if (this.getPref(this.mPrefRoot + "_running"))
			{
				this.delPref(this.mPrefRoot + "_running");
				this.clearUndoData("window", true);
				this.backupCurrentSession();
			}
			break;
		case "domwindowopened":
			aSubject.__gSeMa = this;
			aSubject.addEventListener("load", this.onDomWindowOpened_proxy, false);
			break;
		}
	},

	onDomWindowOpened_proxy: function()
	{
		this.removeEventListener("load", this.__gSeMa.onDomWindowOpened_proxy, false);
		if (this.document.documentElement.getAttribute("windowtype") == "navigator:browser")
		{
			this.__gSeMa.mObserverService.removeObserver(this.__gSeMa, "quit-application");
			this.__gSeMa.mObserverService.removeObserver(this.__gSeMa, "domwindowopened");
		}
		delete this.__gSeMa;
	},

	onTabRestored_proxy: function(aEvent)
	{
		var browser = this.getBrowserForTab(aEvent.originalTarget);
		
		if (gSessionManager.mPref_reload && gSessionManager._allowReload && !browser.__CR_data && !gSessionManager.mIOService.offline)
		{
			browser.reload();
		}
	},

	onToolbarClick: function(aEvent, aButton)
	{
		if (aEvent.button == 1)
		{
			eval("var event = { shiftKey: true }; " + aButton.getAttribute("oncommand"));
		}
		else if (aEvent.button == 2 && aButton.getAttribute("disabled") != "true")
		{
			aButton.open = true;
			aEvent.preventDefault();
		}
	},

/* ........ Menu Event Handlers .............. */

	init: function(aPopup, aIsToolbar)
	{
		function get_(a_id) { return aPopup.getElementsByAttribute("_id", a_id)[0] || null; }
		
		var separator = get_("separator");
		var startSep = get_("start-separator");
		
		for (var item = startSep.nextSibling; item != separator; item = startSep.nextSibling)
		{
			aPopup.removeChild(item);
		}
		
		this.mSessions = this.getSessions();
		this.mSessions.forEach(function(aSession, aIx) {
			var key = (aIx < 9)?aIx + 1:(aIx == 9)?"0":"";
			var menuitem = document.createElement("menuitem");
			menuitem.setAttribute("label", ((key)?key + ") ":"") + aSession.name);
			menuitem.setAttribute("oncommand", 'gSessionManager.load("' + aSession.fileName + '", (event.shiftKey)?"newwindow":(event.ctrlKey)?"append":"");');
			menuitem.setAttribute("onclick", 'if (event.button == 1) gSessionManager.load("' + aSession.fileName + '", "newwindow");');
			menuitem.setAttribute("accesskey", key);
			aPopup.insertBefore(menuitem, separator);
		});
		separator.hidden = (this.mSessions.length == 0);
		this.setDisabled(separator.nextSibling, separator.hidden);
		this.setDisabled(separator.nextSibling.nextSibling, separator.hidden);
		
		get_("resume").setAttribute("checked", this.doResumeCurrent());
		get_("reload").setAttribute("checked", this.mPref_reload);
		get_("overwrite").setAttribute("checked", this.mPref_overwrite);
		
		var undoMenu = get_("undo-menu");
		while (aPopup.lastChild != undoMenu)
		{
			aPopup.removeChild(aPopup.lastChild);
		}
		
		var undoDisabled = (this.mPref_max_closed_undo == 0 && this.mPref_max_tabs_undo == 0);
		var divertedMenu = aIsToolbar && document.getElementById("sessionmanager-undo");
		var canUndo = !undoDisabled && !divertedMenu && this.initUndo(undoMenu.firstChild);
		
		undoMenu.hidden = undoDisabled || divertedMenu || !this.mPref_submenus;
		undoMenu.previousSibling.hidden = undoMenu.hidden && !canUndo;
		this.setDisabled(undoMenu, !canUndo);
		
		if (!this.mPref_submenus && canUndo)
		{
			for (item = undoMenu.firstChild.firstChild; item; item = item.nextSibling)
			{
				aPopup.appendChild(item.cloneNode(true));
			}
		}
		
		// in case the popup belongs to a collapsed element
		aPopup.style.visibility = "visible";
	},

	uninit: function(aPopup)
	{
		aPopup.style.visibility = "";
		
		this.mSessions = null;
	},

	save: function(aName, aOneWindow)
	{
		var values = { text: this.getFormattedName(content.document.title || "about:blank", new Date()) || (new Date()).toLocaleString() };
		if (!aName)
		{
			if (!this.prompt(this._string("save2_session"), this._string("save_" + ((aOneWindow)?"window":"session") + "_ok"), values, this._string("save_" + ((aOneWindow)?"window":"session")), this._string("save_session_ok2")))
			{
				return;
			}
			aName = values.text;
		}
		if (aName)
		{
			var file = this.getSessionDir(values.name || this.makeFileName(aName), !values.name);
			try
			{
				this.writeFile(file, this.getSessionState(aName, aOneWindow));
			}
			catch (ex)
			{
				this.ioError(ex);
			}
		}
	},

	saveWindow: function(aName)
	{
		this.save(aName, true);
	},

	load: function(aName, aMode)
	{
		var state = this.readFile(this.getSessionDir(aName));
		if (!state)
		{
			this.ioError();
			return;
		}
		var mainWin = window;
		var overwriteTabs = true;
		var tabsToMove = null;
		
		aMode = aMode || "default";
		if (aMode == "startup")
		{
			overwriteTabs = this.isCmdLineEmpty();
			tabsToMove = (!overwriteTabs)?Array.slice(gBrowser.mTabs):null;
		}
		else if (aMode == "append")
		{
			overwriteTabs = false;
		}
		else if (aMode == "newwindow" || !this.mPref_overwrite)
		{
			mainWin = this.openWindow(this.getPref("browser.chromeURL"), "chrome,dialog=no,all");
		}
		else
		{
			this.getBrowserWindows().forEach(function(aWindow) {
				if (aWindow != mainWin) { aWindow.close(); }
			});
			this.appendClosedWindow(this.getSessionState("", true), content.document.title);
		}
		
		setTimeout(function() {
			gSessionManager.restoreSession(state, mainWin, overwriteTabs, true);
			if (tabsToMove)
			{
				var endPos = gBrowser.mTabs.length - 1;
				tabsToMove.forEach(function(aTab) { gBrowser.moveTabTo(aTab, endPos); });
			}
		}, 0);
	},

	rename: function()
	{
		var values = {};
		if (!this.prompt(this._string("rename_session"), this._string("rename_session_ok"), values, this._string("rename2_session")))
		{
			return;
		}
		var file = this.getSessionDir(values.name);
		var filename = this.makeFileName(values.text);
		var newFile = (filename != file.leafName)?this.getSessionDir(filename, true):null;
		
		try
		{
			this.writeFile(newFile || file, this.nameState(this.readFile(file), values.text));
			if (newFile)
			{
				this.delFile(file);
			}
		}
		catch (ex)
		{
			this.ioError(ex);
		}
	},

	remove: function(aSession)
	{
		if (!aSession)
		{
			aSession = this.selectSession(this._string("remove_session"), this._string("remove_session_ok"), { multiSelect: true });
		}
		if (aSession)
		{
			aSession.split("\n").forEach(function(aName) {
				gSessionManager.delFile(gSessionManager.getSessionDir(aName));
			});
		}
	},

	openFolder: function()
	{
		try
		{
			this.getSessionDir().launch();
		}
		catch (ex)
		{
			this.ioError(ex);
		}
	},

	toggleResume: function()
	{
		this.setPref("extensions.crashrecovery.resume_session", !this.doResumeCurrent());
	},

	toggleReload: function()
	{
		this.setPref(this.mPrefRoot + "reload", !this.mPref_reload);
	},

	toggleOverwrite: function()
	{
		this.setPref(this.mPrefRoot + "overwrite", !this.mPref_overwrite);
	},

/* ........ Undo Menu Event Handlers .............. */

	initUndo: function(aPopup, aStandAlone)
	{
		function get_(a_id) { return aPopup.getElementsByAttribute("_id", a_id)[0] || null; }
		
		var separator = get_("closed-separator");
		var label = get_("windows");
		
		for (var item = separator.previousSibling; item != label; item = separator.previousSibling)
		{
			aPopup.removeChild(item);
		}
		
		var closedWindows = this.getClosedWindows();
		closedWindows.forEach(function(aWindow, aIx) {
			var menuitem = document.createElement("menuitem");
			menuitem.setAttribute("label", aWindow.name);
			menuitem.setAttribute("oncommand", 'gSessionManager.undoCloseWindow(' + aIx + ', event.shiftKey && event.ctrlKey);');
			aPopup.insertBefore(menuitem, separator);
		});
		label.hidden = (closedWindows.length == 0);
		
		while (aPopup.lastChild != separator.nextSibling)
		{
			aPopup.removeChild(aPopup.lastChild);
		}
		
		this.mClosedTabs.forEach(function(aTab, aIx) {
			var menuitem = document.createElement("menuitem");
			menuitem.setAttribute("label", aTab.name);
			menuitem.setAttribute("oncommand", 'gSessionManager.undoCloseTab(' + aIx + ');');
			aPopup.appendChild(menuitem);
		});
		separator.nextSibling.hidden = (this.mClosedTabs.length == 0);
		separator.hidden = separator.nextSibling.hidden || label.hidden;
		
		var showPopup = closedWindows.length + this.mClosedTabs.length > 0;
		
		if (aStandAlone)
		{
			if (!showPopup)
			{
				this.updateToolbarButton(false);
				setTimeout(function(aPopup) { aPopup.parentNode.open = false; }, 0, aPopup);
			}
			else
			{
				aPopup.style.visibility = "visible";
			}
		}
		
		return showPopup;
	},

	uninitUndo: function(aPopup)
	{
		aPopup.style.visibility = "";
	},

	undoCloseWindow: function(aIx, aOverwrite)
	{
		var closedWindows = this.getClosedWindows();
		if (closedWindows[aIx || 0])
		{
			var state = closedWindows.splice(aIx || 0, 1)[0].state;
			this.storeClosedWindows(closedWindows);
			
			var mainWin = window;
			if (!aOverwrite)
			{
				mainWin = this.openWindow(this.getPref("browser.chromeURL"), "chrome,dialog=no,all");
			}
			else
			{
				this.appendClosedWindow(this.getSessionState("", true), content.document.title);
			}
			
			this.restoreSession(state, mainWin, true);
		}
	},

	undoCloseTab: function(aIx)
	{
		if (this.mClosedTabs[aIx || 0])
		{
			var tabData = this.mClosedTabs.splice(aIx || 0, 1)[0];
			this.restoreSession(tabData.state);
			if (this.mClosedTabs.length == 0)
			{
				this.updateToolbarButton();
			}
			
			var newTab = document.getAnonymousElementByAttribute(gBrowser, "linkedpanel", gBrowser.mPanelContainer.lastChild.id);
			if (tabData.pos && tabData.pos < gBrowser.mTabs.length)
			{ // Firefox and Flock
				gBrowser.moveTabTo(newTab, tabData.pos - 1);
			}
			else if (gBrowser.getTabIndex && gBrowser.selectedTab.nextSibling != newTab)
			{ // SeaMonkey
				gBrowser.moveTabTo(newTab, gBrowser.getTabIndex(gBrowser.selectedTab));
			}
			gBrowser.selectedTab = newTab;
		}
	},

	clearUndoList: function()
	{
		var max_tabs_undo = this.getPref(this.mPrefRoot + "max_tabs_undo", 10);
		
		this.setPref(this.mPrefRoot + "max_tabs_undo", 0);
		this.setPref(this.mPrefRoot + "max_tabs_undo", max_tabs_undo);
		
		this.clearUndoData("window");
	},

/* ........ User Prompts .............. */

	prompt: function(aSessionLabel, aAcceptLabel, aValues, aTextLabel, aAcceptExistingLabel)
	{
		var params = Components.classes["@mozilla.org/embedcomp/dialogparam;1"].createInstance(Components.interfaces.nsIDialogParamBlock);
		
		params.SetNumberStrings(7);
		params.SetString(1, aSessionLabel);
		params.SetString(2, aAcceptLabel);
		params.SetString(3, aValues.name || "");
		params.SetString(4, aTextLabel || "");
		params.SetString(5, aAcceptExistingLabel || "");
		params.SetString(6, aValues.text || "");
		params.SetInt(1, ((aValues.addCurrentSession)?1:0) | ((aValues.multiSelect)?2:0));
		
		this.openWindow("chrome://sessionmanager/content/session_prompt.xul", "chrome,modal,centerscreen,titlebar,resizable", params, (this.mFullyLoaded)?window:null);
		
		aValues.name = params.GetString(3);
		aValues.text = params.GetString(6);
		
		return !params.GetInt(0);
	},

	selectSession: function(aSessionLabel, aAcceptLabel, aValues)
	{
		var values = aValues || {};
		
		if (this.prompt(aSessionLabel, aAcceptLabel, values))
		{
			return values.name;
		}
		
		return null;
	},

	ioError: function(aException)
	{
		this.mPromptService.alert(window, this.mTitle, (this.mBundle)?this.mBundle.getFormattedString("io_error", [(aException)?aException.message:this._string("unknown_error")]):aException);
	},

	openWindow: function(aChromeURL, aFeatures, aArgument, aParent)
	{
		if (!aArgument || typeof(aArgument) == "string")
		{
			var argString = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
			argString.data = aArgument || "";
			aArgument = argString;
		}
		
		return Components.classes["@mozilla.org/embedcomp/window-watcher;1"].getService(Components.interfaces.nsIWindowWatcher).openWindow(aParent || null, aChromeURL, "_blank", aFeatures, aArgument);
	},

/* ........ File Handling .............. */

	getProfileFile: function(aName)
	{
		var file = this.mProfileDirectory.clone();
		file.append(aName);
		return file;
	},

	getSessionDir: function(aName, aUnique)
	{
		var dir = this.getProfileFile("sessions");
		if (!dir.exists())
		{
			dir.create(this.mComponents.interfaces.nsIFile.DIRECTORY_TYPE, 0700);
		}
		if (aName)
		{
			dir.append(aName);
			if (aUnique)
			{
				var postfix = 1, ext = "";
				if (aName.slice(-this.mSessionExt.length) == this.mSessionExt)
				{
					aName = aName.slice(0, -this.mSessionExt.length);
					ext = this.mSessionExt;
				}
				while (dir.exists())
				{
					dir = dir.parent;
					dir.append(aName + "-" + (++postfix) + ext);
				}
			}
		}
		return dir.QueryInterface(this.mComponents.interfaces.nsILocalFile);
	},

	getSessions: function()
	{
		var sessions = [];
		
		var filesEnum = this.getSessionDir().directoryEntries.QueryInterface(Components.interfaces.nsISimpleEnumerator);
		while (filesEnum.hasMoreElements())
		{
			var file = filesEnum.getNext().QueryInterface(Components.interfaces.nsIFile);
			var content = this.readFile(file);
			
			if (gSessionManagerCompatibilityShim.CR_isOldFormat(content))
			{
				content = gSessionManagerCompatibilityShim.encode(gSessionManagerCompatibilityShim.CR_decodeOldFormat(content));
				this.writeFile(file, content.replace(/\n\[/g, "\n$&") + "\n");
			}
			
			if (/^\[SessionManager\]\n(?:name=(.*))?/m.test(content))
			{
				sessions.push({ fileName: file.leafName, name: RegExp.$1 });
			}
		}
		
		return sessions.sort(function(a, b) { return a.name.toLowerCase().localeCompare(b.name.toLowerCase()); });
	},

	getClosedWindows: function()
	{
		var history = this.getProfileFile(this.mClosedWindowFile);
		var data = this.readFile(history);
		
		return (data)?data.split("\n\n").map(function(aEntry) {
			var parts = aEntry.split("\n");
			return { name: parts.shift(), state: parts.join("\n") };
		}):[];
	},

	storeClosedWindows: function(aList)
	{
		this.writeFile(this.getProfileFile(this.mClosedWindowFile), aList.map(function(aEntry) {
			return aEntry.name + "\n" + aEntry.state;
		}).join("\n\n"));
		
		this.updateToolbarButton(aList.length + this.mClosedTabs.length > 0);
	},

	appendClosedWindow: function(aState, aName)
	{
		if (this.mPref_max_closed_undo > 0)
		{
			aState = aState.replace(/^\n+|\n+$/g, "").replace(/\n{2,}/g, "\n");
			var windows = this.getClosedWindows();
			windows.unshift({ name: aName || this._string("untitled_window"), state: aState });
			this.storeClosedWindows(windows.slice(0, this.mPref_max_closed_undo));
		}
	},

	clearUndoData: function(aType, aSilent)
	{
		if (aType == "window" || aType == "all")
		{
			this.delFile(this.getProfileFile(this.mClosedWindowFile), aSilent);
		}
		if (aType == "tab" || aType == "all")
		{
			this.mClosedTabs = [];
		}
		this.updateToolbarButton(false);
	},

	backupCurrentSession: function()
	{
		var backup = this.getPref(this.mPrefRoot + "backup_session", 1);
		if (backup == 2)
		{
			backup = (this.mPromptService.confirmEx(null, this.mTitle, this._string_preserve_session || this._string("preserve_session"), this.mPromptService.BUTTON_TITLE_YES * this.mPromptService.BUTTON_POS_0 + this.mPromptService.BUTTON_TITLE_NO * this.mPromptService.BUTTON_POS_1, null, null, null, null, {}) == 1)?-1:1;
		}
		if (backup > 0)
		{
			try
			{
				var state = this.getSessionState(this._string_backup_session || this._string("backup_session"));
				this.writeFile(this.getSessionDir("backup" + this.mSessionExt), state);
			}
			catch (ex)
			{
				this.ioError(ex);
			}
		}
		else
		{
			if (backup < 0 || !this.doResumeCurrent())
			{
				this.delFile(this.getProfileFile("crashrecovery.dat"), true);
			}
			this.delFile(this.getProfileFile("crashrecovery.bak"), true);
			this.delFile(this.getSessionDir("backup" + this.mSessionExt), true);
		}
	},

	readFile: function(aFile)
	{
		try
		{
			var stream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
			stream.init(aFile, 0x01, 0, 0);
			var sstream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
			sstream.init(stream);
			
			var data = sstream.read(-1).replace(/\r?\n/g, "\n");
			sstream.close();
			stream.close();
			
			return this.mUnicodeConverter.ConvertToUnicode(data);
		}
		catch (ex) { }
		
		return null;
	},

	writeFile: function(aFile, aData)
	{
		if (!aFile.exists())
		{
			aFile.create(this.mComponents.interfaces.nsIFile.NORMAL_FILE_TYPE, 0600);
		}
		var stream = this.mComponents.classes["@mozilla.org/network/file-output-stream;1"].createInstance(this.mComponents.interfaces.nsIFileOutputStream);
		aData = this.mUnicodeConverter.ConvertFromUnicode(aData).replace(/\n/g, this.mEOL);
		
		stream.init(aFile, 0x02 | 0x08 | 0x20, 0600, 0);
		stream.write(aData, aData.length);
		stream.flush();
		stream.close();
	},

	delFile: function(aFile, aSilent)
	{
		if (aFile.exists())
		{
			try
			{
				aFile.remove(false);
			}
			catch (ex)
			{
				if (!aSilent)
				{
					this.ioError(ex);
				}
			}
		}
	},

/* ........ Preference Access .............. */

	getPref: function(aName, aDefault)
	{
		var pb = this.mPrefBranch;
		switch (pb.getPrefType(aName))
		{
		case pb.PREF_STRING:
			return pb.getCharPref(aName);
		case pb.PREF_BOOL:
			return pb.getBoolPref(aName);
		case pb.PREF_INT:
			return pb.getIntPref(aName);
		default:
			return aDefault;
		}
	},

	setPref: function(aName, aValue)
	{
		switch (typeof(aValue))
		{
		case "boolean":
			this.mPrefBranch.setBoolPref(aName, aValue);
			break;
		case "number":
			this.mPrefBranch.setIntPref(aName, parseInt(aValue));
			break;
		default:
			this.mPrefBranch.setCharPref(aName, "" + aValue);
			break;
		}
	},

	delPref: function(aName)
	{
		this.mPrefBranch.deleteBranch(aName);
	},

/* ........ Miscellaneous Enhancements .............. */

	recoverSession: function(aRecoverOnly)
	{
		var recovering = this.getPref(this.mPrefRoot + "_recovering");
		if (recovering)
		{
			this.delPref(this.mPrefRoot + "_recovering");
			this.load(recovering, "startup");
		}
		else if (this.mPref_resume_session && !aRecoverOnly && !this.doResumeCurrent() && this.getSessions().length > 0)
		{
			var session = (this.mPref_resume_session == this.mPromptSessionName)?this.selectSession(this._string("resume_session"), this._string("resume_session_ok")):this.mPref_resume_session;
			if (session && this.getSessionDir(session).exists())
			{
				this.load(session, "startup");
			}
		}
	},

	isCmdLineEmpty: function()
	{
		var homepage = null;
		
		switch (this.getPref("browser.startup.page", 1))
		{
		case 0:
			homepage = "about:blank";
			break;
		case 1:
			try
			{
				homepage = this.mPrefBranch.getComplexValue("browser.startup.homepage", Components.interfaces.nsIPrefLocalizedString).data;
			}
			catch (ex)
			{
				homepage = this.getPref("browser.startup.homepage", "");
			}
			break;
		}
		if (window.arguments.length > 0 && window.arguments[0].split("\n")[0] == homepage)
		{
			window.arguments.shift();
		}
		
		return (window.arguments.length == 0);
	},

	patchTabBrowser: function()
	{
		if (this.mPref_max_tabs_undo > 0 && !gBrowser.__preSM_removeTab)
		{
			gBrowser.__preSM_removeTab = gBrowser.removeTab;
			gBrowser.removeTab = function(aTab) {
				if (aTab.localName != "tab")
				{
					aTab = this.mCurrentTab;
				}
				if (!gSessionManager._ignoreRemovedTabs && !gSessionManager.isCleanBrowser(this.getBrowserForTab(aTab)))
				{
					var preState = gSessionManager.getSessionState();
					var position = (typeof aTab._tPos != "undefined")?aTab._tPos + 1:0;
					var title = aTab.label;
				}
				
				this.__preSM_removeTab(aTab);
				
				if (preState)
				{
					gSessionManager.appendClosedTab(preState, position, title);
				}
			};
			if (!gBrowser.undoRemoveTab) // tweak for Tab Clicking Options
			{
				gBrowser.undoRemoveTab = function() { gSessionManager.undoCloseTab(); };
			}
		}
	},

	updateToolbarButton: function(aEnable)
	{
		var button = (document)?document.getElementById("sessionmanager-undo"):null;
		if (button)
		{
			this.setDisabled(button, (typeof aEnable != "undefined")?!aEnable:this.mClosedTabs.length == 0 && this.getClosedWindows().length == 0);
		}
	},

	appendClosedTab: function(aPreState, aPosition, aTitle)
	{
		if (this.mPref_max_tabs_undo == 0)
		{
			return;
		}
		
		this.mOneTabRE.lastIndex = 0;
		
		var postState = this.getSessionState();
		var match = this.mOneTabRE.exec(aPreState); 
		
		while (postState.indexOf(match[0]) > -1)
		{
			match = this.mOneTabRE.exec(aPreState);
		}
		var tabData = match[0].split(match[1]).join("1.Tab1");
		
		this.mClosedTabs.unshift({ name: aTitle, pos: aPosition, state: tabData });
		this.mClosedTabs.splice(this.mPref_max_tabs_undo);
		
		this.updateToolbarButton(true);
	},

/* ........ Auxiliary Functions .............. */

	getSessionState: function(aName, aOneWindow)
	{
		var state = this.nameState(("[SessionManager]\nname=" + (new Date()).toString() + "\n" + this.mCrashRecoveryService.getCurrentState() + "\n").replace(/\n\[/g, "\n$&"), aName || "");
		
		return (!aOneWindow)?state:state.split("\n[Window2]\n")[0];
	},

	restoreSession: function(aState, aWindow, aReplaceTabs, aAllowReload)
	{
		this._allowReload = aAllowReload;
		this._ignoreRemovedTabs = true;
		this.mCrashRecoveryService.restoreWindow(aWindow || window, aState, aReplaceTabs || false);
		delete this._ignoreRemovedTabs;
	},

	nameState: function(aState, aName)
	{
		if (!/^\[SessionManager\]/m.test(aState))
		{
			return "[SessionManager]\nname=" + aName + "\n\n" + aState;
		}
		return aState.replace(/^(\[SessionManager\])(?:\nname=.*)?/m, function($0, $1) { return $1 + "\nname=" + aName; });
	},

	getFormattedName: function(aTitle, aDate, aFormat)
	{
		function cut(aString, aLength)
		{
			return aString.replace(new RegExp("^(.{" + (aLength - 3) + "}).{4,}$"), "$1...");
		}
		function toISO8601(aDate)
		{
			return [aDate.getFullYear(), pad2(aDate.getMonth() + 1), pad2(aDate.getDate())].join("-");
		}
		function pad2(a) { return (a < 10)?"0" + a:a; }
		
		return (aFormat || this.mPref_name_format).split("%%").map(function(aPiece) {
			return aPiece.replace(/%(\d*)([tdm])/g, function($0, $1, $2) {
				$0 = ($2 == "t")?aTitle:($2 == "d")?toISO8601(aDate):pad2(aDate.getHours()) + ":" + pad2(aDate.getMinutes());
				return ($1)?cut($0, Math.max(parseInt($1), 3)):$0;
			});
		}).join("%");
	},

	makeFileName: function(aString)
	{
		return aString.replace(/[^\w ',;!()@&*+=~\x80-\xFE-]/g, "_").substr(0, 64) + this.mSessionExt;
	},

	getBrowserWindows: function()
	{
		var windowsEnum = this.mWindowMediator.getEnumerator("navigator:browser");
		var windows = [];
		
		while (windowsEnum.hasMoreElements())
		{
			windows.push(windowsEnum.getNext());
		}
		
		return windows;
	},

	doResumeCurrent: function()
	{
		return this.getPref("extensions.crashrecovery.resume_session", false);
	},

	isCleanBrowser: function(aBrowser)
	{
		return aBrowser.sessionHistory.count < 2 && aBrowser.currentURI.spec == "about:blank";
	},

	setDisabled: function(aObj, aValue)
	{
		if (aValue)
		{
			aObj.setAttribute("disabled", "true");
		}
		else
		{
			aObj.removeAttribute("disabled");
		}
	},

	getEOL: function()
	{
		return /win|os[\/_]?2/i.test(navigator.platform)?"\r\n":/mac/i.test(navigator.platform)?"\r":"\n";
	},

	_string: function(aName)
	{
		return this.mBundle.getString(aName);
	}
};

/*const*/ var gSessionManagerCompatibilityShim = {
	encode: function(aObj, aPrefix)
	{
		aPrefix = (aPrefix)?aPrefix + ".":"";
		
		var ini = [], iniChildren = [];
		
		for (var key in aObj)
		{
			if (!key || key.charAt(0) == ";" || key.charAt(0) == "[" || key.indexOf("]") > -1 || key.indexOf("=") > -1)
			{
				// debug("Ignoring invalid key name: '" + key + "'!");
				continue;
			}
			else if (key.charAt(0) == "_")
			{
				continue;
			}
			
			var value = aObj[key];
			if (typeof value == "boolean" || typeof value == "number")
			{
				ini.push(key + "=" + value);
			}
			else if (typeof value == "string" && value)
			{
				ini.push(key + "=" + ((/[%\t\r\n]/.test(value))?encodeURI(value):value));
			}
			else if (value instanceof Array)
			{
				for (var i = 0; i < value.length; i++)
				{
					if (value[i] instanceof Object)
					{
						iniChildren.push("[" + aPrefix + key + (i + 1) + "]");
						iniChildren.push(this.encode(value[i], aPrefix + key + (i + 1)));
					}
				}
			}
			else if (typeof value == "object" && value)
			{
				iniChildren.push("[" + aPrefix + key + "]");
				iniChildren.push(this.encode(value, aPrefix + key));
			}
		}
		
		return ini.concat(iniChildren).join("\n");
	},

	CR_isOldFormat: function(aString)
	{
		return /^.*~~~~/.test(aString);
	},

	CR_decodeOldFormat: function(aString)
	{
		aString = aString.replace(/^(.*?) ~~~~(?: (.*))?/, "$1");
		var root = { SessionManager: { name: RegExp.$2 || gSessionManager._string("untitled_window") } };
		
		root.Window = aString.split("\n- WINDOW -----\n").map(function(aWindow) {
			var winData = {};
			var tabs = aWindow.split("\n-- TAB ----\n");
			var cookies = tabs.shift().split("\n");
			var features = cookies.shift();
			
			winData.selected = parseInt(features) + 1;
			if (/ dims:(\S+)/.test(features))
			{
				RegExp.$1.split(",").forEach(function(aFeat) {
					if (/^(\S+)=(.*)/.test(aFeat)) winData[RegExp.$1] = decodeURI(RegExp.$2);
				});
			}
			if (/ hide:(\S+)/.test(features)) winData.hidden = RegExp.$1;
			winData.sidebar = (/ sidebar:(\S+)/.test(features))?RegExp.$1:"";
			
			cookies.forEach(function(aCookie) {
				if (!window.Cookies) window.Cookies = { count: 0 };
				var cookie = aCookie.split(" ");
				window.Cookies["domain" + ++window.Cookies.count] = cookie[0];
				window.Cookies["value" + window.Cookies.count] = decodeURI(cookie[1]);
			});
			
			winData.Tab = tabs.map(gSessionManagerCompatibilityShim.CR_decodeOldFormatTab);
			
			return winData;
		});
		
		return root;
	},

	CR_decodeOldFormatTab: function(aTab)
	{
		function parseEntry(aTabs, aIx, aPrefix)
		{
			aPrefix = aPrefix || "";
			var entries = aTabs[aIx].split(" ");
			var entry = {
				uri: entries[0].substr(aPrefix.length), scroll: entries[1],
				postdata: decodeURI(entries[2] || ""), title: decodeURI(entries[3] || ""),
				subframe: (entries[4] || "") == "1", Child: []
			};
			
			aPrefix += "+";
			while (aTabs[aIx + 1] && aTabs[aIx + 1].substr(0, aPrefix.length) == aPrefix)
			{
				entry.Child.push(parseEntry(aTabs, aIx + 1, aPrefix));
				aTabs.splice(aIx + 1, 1);
			}
			
			return entry;
		}
		
		var tabData = { Entry: [], index: 0 };
		var tabs = aTab.split("\n");
		if (/^(\d+) zoom:(\S+)/.test(tabs.shift()))
		{
			tabData.index = parseInt(RegExp.$1) + 1;
			tabData.zoom = RegExp.$2;
		}
		
		for (var i = 0; i < tabs.length; i++)
		{
			if (tabs[i].charAt(0) == "@") break;
			tabData.Entry.push(parseEntry(tabs, i));
		}
		for (; i < tabs.length; i++)
		{
			if (/^@(text|framescroll|xulattr|disallow) (.*)/.test(tabs[i]))
			{
				tabData[RegExp.$1] = RegExp.$2;
			}
		}
		
		return tabData;
	}
};

window.addEventListener("load", gSessionManager.onLoad_proxy, false);
window.addEventListener("unload", gSessionManager.onUnload_proxy, false);

window.addEventListener("load", function() {
	if (!window.SessionManager) // if Tab Mix Plus isn't installed
	{
		window.SessionManager = gSessionManager;
	}
	if (typeof window.tabBarScrollStatus == "function") // hack for some Tab Mix Plus startup issue
	{
		setTimeout(window.tabBarScrollStatus, 0);
	}
}, false);
